import dateutil

from crosspm.contracts.contract import Contract
from crosspm.contracts.package_version import PackageVersion
from crosspm.package_parsers.debian_package_name_parser import DebianPackageNameParser


class Package:

    def __init__(self, name: str, version: str, contracts: dict):
        self.name = name
        self.version = PackageVersion(version)
        self.contracts = contracts

    def __hash__(self):
        return hash(str(self.name) + str(self.version))

    def __str__(self):
        return "{}.{}({})".format(self.name, self.version, ",".join(str(s[1]) for s in sorted(self.contracts.items())))

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.name == other.name and self.version == other.version

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return (self.name, self.version) < (other.name, other.version)

    def is_microservice(self, microservice):
        return self.name == microservice

    def is_contract_lower_then(self, other):
        if other.name in self.contracts:
            return self.contracts[other.name].value < other.value

        raise BaseException("package {} has no contract {}", str(self), str(other))

    def is_contract_higher_then(self, other):
        if other.name in self.contracts:
            return self.contracts[other.name].value > other.value

        raise BaseException("package {} has no contract {}", str(self), str(other))

    def is_any_contract_higher(self, other):
        for c in self.calc_contracts_intersection(other.contracts):
            if self.contracts[c].value > other.contracts[c].value:
                return True

        return False

    def calc_contracts_intersection(self, contracts):
        return self.contracts.keys() & contracts.keys()

    @staticmethod
    def create_contracts(contracts):
        return {c[0]: Contract(c[0], c[1]) for c in contracts}

    @staticmethod
    def create_contracts_from_dict(contracts):
        return {k: Contract(k, v) for k, v in contracts.items()}

    @staticmethod
    def create_package(package):

        if len(package) == 3:
            return Package(package[0], str(package[1]), Package.create_contracts(package[2]))

        return Package(package[0], str(package[1]), Package.create_contracts([]))

    @staticmethod
    def create_package_debian(package_name):

        p = DebianPackageNameParser.parse_from_package_name(package_name)

        return Package.create_package((p.package, p.fullversion))

    @staticmethod
    def create_packages(*packages):
        res = set()
        for p in packages:
            res.add(Package.create_package(p))

        return res

class ArtifactoryPackage(Package):
    def __init__(self, name, version, contracts, art_path, params, params_raw, paths_params, arti_item_find_results):
        super(ArtifactoryPackage, self).__init__(name, version,
                                                 Package.create_contracts_from_dict(contracts))
        self.art_path = art_path
        self.params = params
        self.params_raw = params_raw
        self.paths_params = paths_params
        self.arti_item_find_results = arti_item_find_results

    def strisodate_to_timestamp(self, date):
        return dateutil.parser.parse(date).timestamp()

    def pkg_stat(self):
        stat_pkg = {
            'ctime': self.strisodate_to_timestamp(self.arti_item_find_results.created),
            'mtime': self.strisodate_to_timestamp(self.arti_item_find_results.modified),
            'size': int(self.arti_item_find_results.size)
        }
        return stat_pkg




import logging

from ordered_set import OrderedSet

from crosspm.helpers.exceptions import CrosspmException, CROSSPM_ERRORCODE_PACKAGE_NOT_FOUND


class Bundle:
    def __init__(self, deps, packages_repo, trigger_package):
        # it is vital for deps to be list, orderedset (or something with insertion order savings),
        # we need the order of packages in dependencies.txt to take next package
        # when no contracts satisfied
        self._log = logging.getLogger('crosspm')

        self._deps = OrderedSet(deps)
        self._packages_repo = sorted(packages_repo, reverse=True)

        self._trigger_package = None
        if trigger_package:
            self._trigger_package = Bundle.find_trigger_package_in_packages_repo(trigger_package, self._packages_repo)

        self._packages = dict()
        self._bundle_contracts = {}

    @staticmethod
    def find_trigger_package_in_packages_repo(trigger_package, repo_packages):

        for p in repo_packages:
            if p == trigger_package:
                return p

        raise CrosspmException(CROSSPM_ERRORCODE_PACKAGE_NOT_FOUND, f"trigger_package = <{trigger_package}> NOT FOUND in repo_packages : {repo_packages}")

    def calculate(self):

        self._log.info('trigger_package: {}'.format(self._trigger_package))
        self._log.info('packages_repo: {}'.format(self._packages_repo))
        if self._trigger_package:
            self._packages[self._trigger_package.name] = self._trigger_package

        while True:
            rest_packages_to_find = self.rest_packages_to_find(self._deps, self._packages)
            if not rest_packages_to_find:
                break

            self.update_bundle_contracts()
            self.find_next_microservice_package(rest_packages_to_find)

        return self._packages

    def find_next_microservice_package(self, rest_packages_to_find):
        next_packages_out_of_current_contracts = dict()
        package_lowering_contract = []

        for m in rest_packages_to_find:
            for p in [i for i in self._packages_repo if i.is_microservice(m)]:
                package = self.is_package_corresponds_bundle_current_contracts(next_packages_out_of_current_contracts, p,
                                                                               self._bundle_contracts, package_lowering_contract)
                if package:
                    self._package_add(package)
                    return

            if package_lowering_contract:
                p = package_lowering_contract[0]

                self.remove_packages_with_higher_contracts_then(p)
                self._package_add(p)

                return

        package = self.select_next_microservice_package_out_of_current_contracts(next_packages_out_of_current_contracts, rest_packages_to_find)
        if package:
            self._package_add(package)
            return

        raise BaseException("cant select next package for current contracts:\n"
                            "next_packages_out_of_current_contracts : {}\n"
                            "rest_packages_to_find : {}"
                            .format(next_packages_out_of_current_contracts, rest_packages_to_find))

    def select_next_microservice_package_out_of_current_contracts(self, next_packages_out_of_current_contracts, select_order):
        for i in select_order:
            if i in next_packages_out_of_current_contracts:
                return next_packages_out_of_current_contracts[i]

        return None

    def is_package_corresponds_bundle_current_contracts(self, next_packages_out_of_current_contracts, package,
                                                        bundle_contracts, package_lowering_contract):

        intersection_package_contracts = package.calc_contracts_intersection(bundle_contracts)

        if not intersection_package_contracts:
            cp = None
            if package.name in next_packages_out_of_current_contracts:
                cp = next_packages_out_of_current_contracts[package.name]

            if cp is None:
                next_packages_out_of_current_contracts[package.name] = package
            elif cp.version < package.version:
                next_packages_out_of_current_contracts[package.name] = package

            return None

        failed_contracts = set(intersection_package_contracts)
        for c in intersection_package_contracts:

            if package.contracts[c] == bundle_contracts[c]:
                failed_contracts.discard(c)

            if package.contracts[c].value < bundle_contracts[c].value:
                if not package_lowering_contract:
                    package_lowering_contract.append(package)
                elif package_lowering_contract[0].is_contract_lower_then(package.contracts[c]):
                    package_lowering_contract.clear();
                    package_lowering_contract.append(package)


        if not failed_contracts:
            return package

    def remove_packages_with_higher_contracts_then(self, package_lowering_contract):

        for p in [*self._packages.values()]:
            if p.is_any_contract_higher(package_lowering_contract):
                if self._trigger_package is not None and self._trigger_package == p:
                    raise BaseException(
                        "no tree resolve with trigger_package {} has no appropriate packages with specified package contracts"
                        .format(self._trigger_package))
                del self._packages[p.name]

    def update_bundle_contracts(self):
        self._bundle_contracts = dict()
        for p in self._packages.values():
            self._bundle_contracts.update(p.contracts)

    def rest_packages_to_find(self, deps, packages):
        return deps - packages.keys()

    def _package_add(self, package):
        self._packages[package.name] = package



class Contract:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __hash__(self):
        return hash((self.name, self.value))

    def __str__(self):
        return "{}.{}".format(self.name, self.value)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.name == other.name and self.value == other.value

    def __ne__(self, other):
        return not (self == other)

class PackageContracts:
    def __init__(self, contracts):
        self._contracts = contracts

    def __getitem__(self, key):
        for c in self._contracts:
            if c.name == key.name:
                return c

        return None

    def is_lower(self, contract):
        c = self[contract]
        return c and c.value < contract.value





# -*- coding: utf-8 -*-
from crosspm.helpers.locker import Locker

from crosspm.helpers.parser2 import Parser2
from crosspm.output_formatters.deps_txt_lock_list_formatter import DepsTxtLockListFormatter


class Locker2(Locker):
    def __init__(self, config, do_load, recursive):
        # TODO: revise logic to allow recursive search without downloading
        super(Locker2, self).__init__(config, do_load, recursive, Parser2)

    def lock_packages(self, deps_file_path=None, depslock_file_path=None, packages=None):
        """
        Lock packages. Downloader search packages
        """
        if deps_file_path is None:
            deps_file_path = self._deps_path
        if depslock_file_path is None:
            depslock_file_path = self._depslock_path
        if deps_file_path == depslock_file_path:
            depslock_file_path += '.lock'

        if packages is None:
            self.search_dependencies(deps_file_path)
        else:
            self._root_package.packages = packages

        self._log.info('Writing lock file [{}]'.format(depslock_file_path))

        # DepsTxtLockTableFormatter.write(depslock_file_path, self._root_package.packages)

        DepsTxtLockListFormatter.write_packages_to_lock_file(depslock_file_path, self._root_package.packages)

        self._log.info('Done!')




# -*- coding: utf-8 -*-
import json
import logging
import os
import platform
import sys

import requests
import yaml

from crosspm.adapters.artifactoryaql import ArtifactoryAql
from crosspm.adapters.artifactoryaql2 import ArtifactoryAql2
from crosspm.helpers.cache import Cache
from crosspm.helpers.content import DependenciesContent
from crosspm.helpers.exceptions import *
from crosspm.helpers.parser import Parser
from crosspm.helpers.parser2 import Parser2
from crosspm.helpers.source import Source
import crosspm.contracts.package

requests.packages.urllib3.disable_warnings()

WINDOWS = (platform.system().lower() == 'windows') or (os.name == 'nt')
DEFAULT_CONFIG_FILE = ('crosspm.yaml', 'crosspm.yml', 'crosspm.json',)
USER_HOME_DIR = os.path.expanduser('~')
CROSSPM_ROOT_DIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
DEFAULT_CONFIG_PATH = [
    './',
    '~/',
    '~/.crosspm',
    '/etc/crosspm',
    CROSSPM_ROOT_DIR,
] if not WINDOWS else [
    './',
    os.path.realpath(os.path.join(os.path.splitdrive(USER_HOME_DIR)[0], '/')),
    os.path.realpath(os.path.join(USER_HOME_DIR, '.crosspm')),
    os.path.realpath(os.path.join(os.getenv('APPDATA'), 'crosspm')),
    CROSSPM_ROOT_DIR,
]
GLOBAL_CONFIG_FILE = ['crosspm_global.yaml', 'global.yaml']
GLOBAL_CONFIG_PATH = [
    '/etc/crosspm',
] if not WINDOWS else [
    '{system_drive}',
    '{all_users}\\crosspm',
    '{program_data}\\crosspm',
]

ENVIRONMENT_CONFIG_PATH = 'CROSSPM_CONFIG_PATH'
CROSSPM_DEPENDENCY_FILENAME = 'dependencies.txt'  # maybe 'cpm.manifest'
CROSSPM_DEPENDENCY_LOCK_FILENAME = CROSSPM_DEPENDENCY_FILENAME  # 'dependencies.txt.lock'
CROSSPM_ADAPTERS_NAME = 'adapters'
CROSSPM_ADAPTERS_DIR = os.path.join(CROSSPM_ROOT_DIR, CROSSPM_ADAPTERS_NAME)


class FactoryParser:
    def __init__(self):
        self.parsers = {}
        self.register_parser('repo', Parser)
        self.register_parser('repo2', Parser2)

    def register_parser(self, name, parser_cls):
        self.parsers[name] = parser_cls

    def create(self, name, data, config):
        parser_cls = self.parsers[name]
        return parser_cls(name, data, config)

#
# class FactoryAdapter:
#     def __init__(self):
#         self.adapters = {}
#         self.register_adapter('artifactory-aql', ArtifactoryAql)
#         self.register_adapter('jfrog-artifactory-aql', ArtifactoryAql)
#         self.register_adapter('artifactory', ArtifactoryAql)
#         self.register_adapter('jfrog-artifactory', ArtifactoryAql)
#         self.register_adapter('jfrog-artifactory-aql2', ArtifactoryAql2)
#
#     def register_adapter(self, name, adapter_cls):
#         self.adapters[name] = adapter_cls
#
#     def create(self, name, data, config):
#         adapter_cls = self.adapters[name]
#         return adapter_cls(name, data, config)


factory_parser = FactoryParser()
# factory_adapter = FactoryAdapter()


class Config:
    windows = WINDOWS

    def __init__(self, config_file_name='', cmdline='', no_fails=False, output_path='', depslock_path='', deps_path='',
                 lock_on_success=False, prefer_local=False, trigger_package=None):
        self._log = logging.getLogger('crosspm')
        self._config_path_env = []
        self._sources = []
        self._adapters = {}
        self._parsers = {}
        self._defaults = {}
        self._columns = []
        self._not_columns = {}
        self._options = {}
        self._values = {}
        self._output = {}
        self._solid = {}
        self._fails = {}
        self.name_column = ''
        self.deps_file_name = ''
        self.deps_lock_file_name = ''
        self.lock_on_success = lock_on_success
        self.prefer_local = prefer_local
        if trigger_package:
            self.trigger_package = crosspm.contracts.package.Package.create_package_debian(trigger_package)
        else:
            self.trigger_package = None
        self.crosspm_cache_root = ''
        self.deps_path = ''
        self.depslock_path = ''
        self.cache_config = {}
        self.init_env_config_path()
        self.secret_variables = []
        self.output_path = output_path
        cpm_conf_name = ''
        if deps_path:
            if deps_path.__class__ is DependenciesContent:
                # HACK
                self.deps_path = deps_path
            else:
                deps_path = deps_path.strip().strip('"').strip("'")
                self.deps_path = os.path.realpath(os.path.expanduser(deps_path))
            if not cpm_conf_name:
                cpm_conf_name = self.get_cpm_conf_name(deps_path)
            if os.path.isfile(deps_path):
                config_path_tmp = os.path.dirname(deps_path)
            else:
                config_path_tmp = deps_path
            if config_path_tmp not in DEFAULT_CONFIG_PATH:
                DEFAULT_CONFIG_PATH.append(config_path_tmp)

        if depslock_path:
            if depslock_path.__class__ is DependenciesContent:
                # HACK
                self.depslock_path = depslock_path
            else:
                depslock_path = depslock_path.strip().strip('"').strip("'")
                self.depslock_path = os.path.realpath(os.path.expanduser(depslock_path))

            if not cpm_conf_name:
                cpm_conf_name = self.get_cpm_conf_name(depslock_path)
            if os.path.isfile(depslock_path):
                config_path_tmp = os.path.dirname(depslock_path)
            else:
                config_path_tmp = depslock_path
            if config_path_tmp not in DEFAULT_CONFIG_PATH:
                DEFAULT_CONFIG_PATH.append(config_path_tmp)

        self._config_file_name = self.find_config_file(config_file_name, cpm_conf_name)
        self._global_config_file_name = self.find_global_config_file()
        if self._global_config_file_name:
            config_data = self.read_config_file(True)
            _override = config_data.pop('override', True)
            if isinstance(_override, str):
                if _override.lower() in ['0', 'no', '-', 'false']:
                    _override = False
            try:
                _override = bool(_override)
            except Exception:
                _override = True
        else:
            config_data = {}
            _override = False
        if _override:
            config_data.update({k: v for k, v in self.read_config_file().items() if k not in config_data})
            self._log.debug('Overriding config file values with global config.')
        else:
            config_data.update(self.read_config_file())

        # Add secret variables to special list in config
        if 'options' in config_data:
            for variable, property in config_data['options'].items():
                if property.get('secret', False):
                    self.secret_variables.append(variable)

        self.no_fails = no_fails
        self.parse_config(config_data, cmdline)
        self.cache = Cache(self, self.cache_config)
        # self._fails = {}

    def get_cpm_conf_name(self, deps_filename=''):
        if not deps_filename:
            deps_filename = self.depslock_path
        result = ''
        if os.path.isfile(deps_filename):
            try:
                with open(deps_filename, 'r', encoding="utf-8-sig") as f:
                    for line in f:
                        line = line.strip()
                        if line.startswith('#'):
                            line = [x.strip().strip('"').strip("'") for x in line.strip('#').split('=') if x]
                            if len(line) > 1:
                                if line[0].lower() == 'cpmconfig':
                                    result = line[1].split('#')[0].strip('"').strip("'")
                                    break
            except Exception:
                pass
        return result

    def init_env_config_path(self):
        self._config_path_env = [x for x in os.getenv(ENVIRONMENT_CONFIG_PATH, '').split(';') if x]
        for config_path_env in self._config_path_env:
            config_path_tmp = os.path.dirname(config_path_env) if os.path.isfile(config_path_env) else config_path_env
            if config_path_tmp in DEFAULT_CONFIG_PATH:
                DEFAULT_CONFIG_PATH.remove(config_path_tmp)
            DEFAULT_CONFIG_PATH.insert(0, config_path_tmp)

    def find_global_config_file(self):
        args = {}
        _win_disk = self.get_windows_system_disk()
        if WINDOWS:
            args = {
                'system_drive': _win_disk,
                'all_users': os.getenv('AllUsersProfile', os.path.join(_win_disk, 'Users\\All Users')),
                'program_data': os.getenv('ProgramData', os.path.join(_win_disk, 'ProgramData')),
            }

        for config_path_env in self._config_path_env:
            if os.path.isdir(config_path_env):
                if config_path_env in GLOBAL_CONFIG_PATH:
                    GLOBAL_CONFIG_PATH.remove(config_path_env)
                GLOBAL_CONFIG_PATH.insert(0, config_path_env)

        for config_path in GLOBAL_CONFIG_PATH:
            try:
                _path = config_path.format(**args).strip().strip("'").strip('"')
            except Exception:
                _path = ''
            if _path:
                for config_name in GLOBAL_CONFIG_FILE:
                    _file = os.path.realpath(os.path.expanduser(os.path.join(_path, config_name)))
                    if os.path.isfile(_file):
                        return _file
        return ''

    @staticmethod
    def find_cpmconfig(conf_name=''):
        conf_path_add = ''
        try:
            _temp = __import__('cpmconfig.cpmconfig', globals(), locals(), ['App'], 0)
            cpm_find = _temp.App.find
            _temp_conf = __import__('cpmconfig.conf', globals(), locals(), ['conf', 'conf_file', 'conf_path'], 0)
            conf_path = getattr(_temp_conf, 'conf_path', '')
            if not conf_path:
                conf_file = getattr(_temp_conf, 'conf_file', '')
                if not conf_file:
                    conf = getattr(_temp_conf, 'conf', '')
                    if conf:
                        conf_path = cpm_find(None, conf)[2]
                else:
                    conf_path = cpm_find(None, conf_file)[2]
            if conf_name:
                conf_path_add = cpm_find(None, conf_name)[2]

        except Exception:
            conf_path = ''
        if conf_path and not os.path.isfile(conf_path):
            conf_path = ''
        if conf_path_add and not os.path.isfile(conf_path_add):
            conf_path_add = ''
        return conf_path, conf_path_add

    def find_config_file(self, config_file_name='', cpm_conf_name=''):
        if not config_file_name:
            ind = 0
            for config_path_env in self._config_path_env:
                if os.path.isfile(config_path_env):
                    if config_path_env in DEFAULT_CONFIG_PATH:
                        DEFAULT_CONFIG_PATH.remove(config_path_env)
                    DEFAULT_CONFIG_PATH.insert(ind, config_path_env)
                    ind += 1

            cpm_file_default, cpm_file_deps = self.find_cpmconfig(cpm_conf_name)
            if cpm_file_deps:
                DEFAULT_CONFIG_PATH.insert(ind, cpm_file_deps)

            _def_conf_file = [DEFAULT_CONFIG_FILE] if isinstance(DEFAULT_CONFIG_FILE, str) else DEFAULT_CONFIG_FILE
            for config_path in DEFAULT_CONFIG_PATH:
                config_path = config_path.strip().strip("'").strip('"')
                config_path = os.path.realpath(os.path.expanduser(config_path))
                for _conf_file in _def_conf_file:
                    config_file_name = os.path.realpath(os.path.join(config_path, _conf_file)) if os.path.isdir(
                        config_path) else config_path
                    if os.path.isfile(config_file_name):
                        break
                    else:
                        config_file_name = ''
                if config_file_name:
                    break

            if not config_file_name:
                config_file_name = cpm_file_default
            if config_file_name:
                self._log.info('Found config file [%s]', config_file_name)
        else:
            config_file_name = config_file_name.strip().strip("'").strip('"')
            config_file_name = os.path.realpath(os.path.expanduser(config_file_name))
            self._log.info('Check config file [%s]', config_file_name)
            if not os.path.isfile(config_file_name):
                config_file_name = ''
            else:
                config_path_tmp = os.path.dirname(config_file_name)
                if config_path_tmp not in DEFAULT_CONFIG_PATH:
                    DEFAULT_CONFIG_PATH.append(config_path_tmp)

        if config_file_name == '':
            raise CrosspmException(
                CROSSPM_ERRORCODE_CONFIG_NOT_FOUND,
                'Config file does not found',
            )

        return config_file_name

    def read_config_file(self, _global=False):
        if _global:
            _config_file_name = self._global_config_file_name
            self._log.info('Reading global config file...')
        else:
            _config_file_name = self._config_file_name
            self._log.info('Reading config file... [%s]', _config_file_name)
        _ext = os.path.splitext(_config_file_name)[1].lower()
        _is_yaml = True
        if _ext in ['.yaml', '.yml']:
            _is_yaml = True
        elif _ext == '.json':
            _is_yaml = False
        else:
            with open(_config_file_name, encoding="utf-8-sig") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    if line.startswith('{'):
                        _is_yaml = False  # Assuming JSON format
                        break

        try:
            if _is_yaml:
                result = self.load_yaml(_config_file_name)
            else:
                with open(_config_file_name, encoding="utf-8-sig") as f:
                    result = json.loads(f.read())

        except Exception as e:
            self._log.exception(e)
            code = CROSSPM_ERRORCODE_CONFIG_IO_ERROR
            msg = 'Error reading config file: [{}]'.format(self._config_file_name)
            raise CrosspmException(code, msg) from e

        return result

    @staticmethod
    def find_import_file(import_file_name=''):
        res_file_name = ''
        if import_file_name:
            for config_path in [''] + DEFAULT_CONFIG_PATH:
                if config_path:
                    config_path = config_path.strip().strip("'").strip('"')
                    config_path = os.path.realpath(os.path.expanduser(config_path))
                if config_path == '' or os.path.isdir(config_path):
                    res_file_name = os.path.realpath(
                        os.path.expanduser(import_file_name) if config_path == '' else
                        os.path.join(config_path, import_file_name))
                    if os.path.isfile(res_file_name):
                        break
                    else:
                        res_file_name = ''
                else:
                    res_file_name = ''

            if res_file_name == '':
                raise CrosspmException(
                    CROSSPM_ERRORCODE_CONFIG_NOT_FOUND,
                    'Config import file "{}" does not found'.format(import_file_name),
                )

        return res_file_name

    @staticmethod
    def load_yaml(_config_file_name):
        result = {}
        yaml_imports = ''
        yaml_content = ''
        with open(_config_file_name, encoding="utf-8-sig") as f:
            for line in f:
                if line.strip().startswith('#'):
                    continue
                if (not yaml_imports) and (not yaml_content):
                    line_one = line.strip().replace(' ', '').lower()
                    if line_one.startswith('import:'):
                        yaml_imports += line
                    elif len(line_one) > 0:
                        yaml_content += line
                elif yaml_imports and not yaml_content:
                    if line.startswith((' ', '\t',)):
                        yaml_imports += line
                    elif line.strip():
                        yaml_content += line
                else:
                    yaml_content += line
        if yaml_imports:
            data_imports = yaml.safe_load(yaml_imports)
            yaml_content_pre = ''
            if 'import' in data_imports:
                if isinstance(data_imports['import'], str):
                    data_imports['import'] = [data_imports['import']]
                if isinstance(data_imports['import'], (list, tuple)):
                    for _import_file_name in data_imports['import']:
                        _import_file_name = Config.find_import_file(_import_file_name)
                        if _import_file_name:
                            with open(_import_file_name, encoding="utf-8-sig") as f:
                                for line in f:
                                    yaml_content_pre += line
                            yaml_content_pre += '\n'
            yaml_content = yaml_content_pre + yaml_content
        if yaml_content:
            result = yaml.safe_load(yaml_content)
        return result

    def parse_config(self, config_data, cmdline):
        # init package_parsers
        if 'parsers' in config_data:
            self.init_parsers(config_data['parsers'])

        # init columns
        if 'columns' in config_data:
            self._columns = []
            _name_index = 0
            for i, _col in enumerate([x for x in [x.strip().split(' ') for x in config_data['columns'].split(',')]]):
                if _col[0][0] == '*':
                    _name_index = i
                    _col = [_col[0][1:]]
                self._columns += _col
            self.name_column = self._columns[_name_index]

        # init available values for columns
        if 'values' in config_data:
            self._values = config_data['values']

        # init options
        if 'options' in config_data:
            self._options = config_data['options']

        # init default values for columns
        if 'defaults' in config_data:
            self._defaults = config_data['defaults']
        self._defaults.update(
            {k: v['default'] for k, v in self._options.items() if k not in self._defaults and 'default' in v})

        # init solid
        if 'solid' in config_data:
            self._solid = config_data['solid']
            if 'ext' in self._solid and not isinstance(self._solid['ext'], (list, tuple)):
                self._solid['ext'] = [self._solid['ext']]

        # init fails
        if 'fails' in config_data:
            self._fails = config_data['fails']

        # init output
        if 'output' in config_data:
            self._output = config_data['output']

        # init common parameters
        _common = {}
        if 'common' in config_data:
            _common = config_data['common']

        if 'sources' in config_data:

            # init adapters
            _types = []
            if 'type' in _common:
                _types += [_common['type']]
            for _src in config_data['sources']:
                if 'type' in _src and _src['type'] not in _types:
                    _types += [_src['type']]

            self.init_adapters(_types)

            # init sources
            for _src in config_data['sources']:
                _src.update({k: v for k, v in _common.items() if k not in _src})

                if 'type' in _src:
                    if _src['type'] not in self._adapters.keys():
                        code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                        msg = 'Adapter [{}] does not found!'.format(_src['type'])
                        self._log.exception(msg)
                        raise CrosspmException(code, msg)
                else:
                    code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                    msg = 'Source must have type property!'
                    self._log.exception(msg)
                    raise CrosspmException(code, msg)

                if 'parser' in _src:
                    if _src['parser'] not in self._parsers.keys():
                        code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                        msg = 'Parser [{}] does not found!'.format(_src['parser'])
                        self._log.exception(msg)
                        raise CrosspmException(code, msg)

                self._sources.append(
                    Source(
                        self._adapters[_src['type']],
                        self._parsers[_src['parser']],
                        _src
                    )
                )

        if len(self._sources) == 0:
            code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
            msg = 'No correct sources defined! Unable to process any further.'
            self._log.exception(msg)
            raise CrosspmException(code, msg)

        try:
            _cmdline = {x[0]: x[1] for x in [x.strip().split('=') for x in cmdline.split(',')] if len(x) > 1}
        except Exception:
            _cmdline = {}

        # init cpm parameters
        cpm_init_done = False

        for x in ['cpm', 'crosspm']:
            if x in config_data:
                # init cache:
                if 'cache' not in config_data:
                    config_data['cache'] = {}
                self.init_cpm_and_cache(config_data[x], _cmdline, config_data['cache'])
                cpm_init_done = True
                break
        if not cpm_init_done:
            if 'cache' not in config_data:
                config_data['cache'] = {}
            self.init_cpm_and_cache({}, _cmdline, config_data['cache'])

        self._options = self.parse_options(self._options, _cmdline)

        # init not columns:
        self.init_not_columns()

    def init_cpm_and_cache(self, crosspm, cmdline, cache_config):
        if 'cache' not in crosspm:
            crosspm['cache'] = {}

        # have to merge dicts:
        parts = ('clear', 'storage')
        cache_config.update(
            {k: v for k, v in crosspm['cache'].items() if (k not in cache_config) and (k not in parts)})
        for part in parts:
            if part in crosspm['cache']:
                if part not in cache_config:
                    cache_config[part] = {}
                cache_config[part].update(
                    {k: v for k, v in crosspm['cache'][part].items() if k not in cache_config[part]})

        if 'path' in cache_config:
            crosspm['cache'] = cache_config.pop('path')
        else:
            for x in ['env', 'cmdline', 'default']:
                if x in cache_config:
                    crosspm['cache'][x] = cache_config.pop(x)

        crosspm = self.parse_options(crosspm, cmdline)

        self.deps_lock_file_name = crosspm.get('dependencies-lock', CROSSPM_DEPENDENCY_LOCK_FILENAME)
        if 'dependencies' in crosspm:
            self.deps_file_name = crosspm.get('dependencies', CROSSPM_DEPENDENCY_FILENAME)
            if 'dependencies-lock' not in crosspm:
                self.deps_lock_file_name = self.deps_file_name

        if not self.prefer_local:
            self.prefer_local = crosspm.get('prefer-local', False)
        if not self.lock_on_success:
            lock_on_success = crosspm.get('lock-on-success', False)
            if isinstance(lock_on_success, str):
                if lock_on_success.lower() in ['1', 'yes', '+', 'true']:
                    self.lock_on_success = True
            try:
                self.lock_on_success = bool(lock_on_success)
            except Exception:
                self.lock_on_success = False

        # Cache init
        self.crosspm_cache_root = crosspm.get('cache', '').strip().strip('"').strip("'")
        if not self.crosspm_cache_root:
            home_dir = os.getenv('APPDATA') if WINDOWS else os.getenv('HOME')
            self.crosspm_cache_root = os.path.join(home_dir, '.crosspm')
        self.crosspm_cache_root = os.path.realpath(os.path.expanduser(self.crosspm_cache_root))
        self.cache_config = cache_config

    def init_not_columns(self):
        # gather items from options
        # include options here even if they are columns also
        self._not_columns = {k: v for k, v in self._options.items()}

        # gather items from defaults
        self._not_columns.update({k: v for k, v in self._defaults.items() if k not in self._not_columns})

        # gather items from package_parsers
        for _parser in self._parsers.values():
            self._not_columns.update({k: None for k in _parser.get_vars() if k not in self._not_columns})

    def init_parsers(self, parsers):
        if 'common' not in parsers:
            parsers['common'] = {}
        for k, v in parsers.items():
            if k != 'common':
                if k not in self._parsers:
                    v.update({_k: _v for _k, _v in parsers['common'].items() if _k not in v})

                    self._parsers[k] = factory_parser.create(k, v, self)
                else:
                    code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                    msg = 'Config file contains multiple definitions of the same parser: [{}]'.format(k)
                    self._log.exception(msg)
                    raise CrosspmException(code, msg)
        if len(self._parsers) == 0:
            code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
            msg = 'Config file does not contain package_parsers! Unable to process any further.'
            self._log.exception(msg)
            raise CrosspmException(code, msg)

    def get_parser(self, parser_name):
        return self._parsers[parser_name] if parser_name in self._parsers else None

    def init_adapters(self, types):

        self._adapters['jfrog-artifactory-aql2'] = ArtifactoryAql2(self)
        self._adapters['artifactory-aql'] = ArtifactoryAql(self)
        self._adapters['jfrog-artifactory-aql'] = ArtifactoryAql(self)
        self._adapters['artifactory'] = ArtifactoryAql(self)
        self._adapters['jfrog-artifactory'] = ArtifactoryAql(self)

    @staticmethod
    def parse_options(options, cmdline, check_default=False):
        # try:
        #     _cmdline = {x[0]: x[1] for x in [x.strip().split('=') for x in cmdline.split(',')] if len(x) > 1}
        # except Exception:
        #     _cmdline = {}
        _remove = []
        for k, v in options.items():
            # if option type is str, leave the value intact
            if not isinstance(options[k], (str, bool)):
                # if option has cmdline name, try to fetch a value from command line by cmdline name
                if 'cmdline' in v:
                    if v['cmdline'] in cmdline:
                        options[k] = cmdline[v['cmdline']]

                # if option hasn't cmdline name, try to fetch a value from command line by option's name
                elif k in cmdline:
                    options[k] = cmdline[k]

                # if option is still not filled, try to get a value from environment by env name
                if not isinstance(options[k], str):
                    success = False
                    if 'env' in v:
                        _env = os.getenv(v['env'])
                        if _env:
                            options[k] = _env
                            success = True
                    # if option is still not filled, try to fill with default value if necessary
                    if not success:
                        if check_default and 'default' in v:
                            options[k] = v['default']
                            success = True
                    # if option is still not filled again, just remove it from dict
                    if not success:
                        _remove.append(k)
        for k in _remove:
            options.pop(k)
        return options

    def sources(self):
        for _src in self._sources:
            yield _src

    def check_column_value(self, i, v, get_list=False):
        if i < len(self._columns):
            k = self._columns[i]
            if k in self._options:
                v = self._options[k]
            if v is None:
                if k in self._defaults:
                    v = self._defaults[k]
        else:
            k, v = '', ''
        return [k, v] if get_list else {k: v}

    def get_column_name(self, i):
        return self._columns[i] if len(self._columns) - 1 >= i else None

    def get_columns(self):
        return self._columns

    def complete_params(self, _vars, update=True):
        result = _vars if update else {}
        result.update({k: v for k, v in self._not_columns.items() if k not in _vars})
        result.update({k: '' for k in self._columns if (k not in _vars) and (k not in result)})
        return result

    def get_values(self, column_name):
        _res = None
        if column_name in self._values:
            _res = self._values[column_name]
        return _res

    def iter_valued_columns(self, column_names):
        for column_name in column_names:
            if column_name in self._values:
                yield column_name

    def iter_valued_columns2(self, column_names):
        for column_name in column_names:
            yield column_name, column_name in self._values

    def output(self, out_type, default):
        if out_type in self._output:
            result = self._output[out_type]
        else:
            result = default
        return result

    def get_fails(self, fail_type, default):
        return self._fails.get(fail_type, default)

    def solid(self, package):
        if 'ext' in self._solid:
            if package.ext(self._solid['ext']):
                return True
        return False

    @staticmethod
    def get_verbosity_level(level=None, text=False):
        levels = (
            ('critical', logging.CRITICAL, False),
            ('error', logging.ERROR, False),
            ('warning', logging.WARNING, False),
            ('info', logging.INFO, True),
            ('debug', logging.DEBUG, False),
        )
        if level is None:
            return ', '.join(x[0] for x in levels)

        if level == 'console':
            level = 'debug'

        default = None
        for x in levels:
            if isinstance(level, str):
                if x[0] == level.lower():
                    return x[1]
            elif x[2]:
                default = x[0] if text else x[1]
                break

        return default

    @staticmethod
    def get_windows_system_disk():
        result = os.getenv('SystemDrive', '')
        if not result:
            try:
                _temp = __import__('win32api', globals(), locals(), ['GetSystemDirectory'], 0)
                result = _temp.GetSystemDirectory()
                if result:
                    result = os.path.splitdrive(result)[0]
            except Exception:
                pass
        if not result:
            result = 'C:'
        return result



# -*- coding: utf-8 -*-
from pathlib import PurePath

from addict import Dict

from crosspm.contracts.package_version import PackageVersion
from crosspm.helpers.parser import Parser
from crosspm.package_parsers.debian_package_name_parser import DebianPackageNameParser


class Parser2(Parser):
    def __init__(self, name, data, config):
        # pass
        super().__init__(name, data, config)

    def get_vars(self):
        return ['server', 'repo', 'package', 'version']

    def validate_path(self, path, params):
        p = PurePath(path)
        debian_package = DebianPackageNameParser.parse_from_package_name(p.name)
        package_version = PackageVersion(debian_package.fullversion)

        res_params = Dict()
        res_params.update(params)
        res_params.version = package_version.release

        res_params_raw = Dict()
        res_params_raw.version = debian_package.fullversion

        return True, res_params, res_params_raw

    def merge_with_mask(self, column, value):
        return value




# -*- coding: utf-8 -*-
# from crosspm.helpers.exceptions import *
import itertools

from addict import Dict
from requests.auth import HTTPBasicAuth


class Source:
    def __init__(self, adapter, parser, data):
        self._adapter = adapter
        self._parser = parser
        if 'repo' not in data:
            data['repo'] = ''
        if isinstance(data['repo'], str):
            data['repo'] = [data['repo']]
        self.args = Dict({k: v for k, v in data.items() if k not in ['type', 'parser']})
        self.path_patterns = parser._rules['path']

    @property
    def repos(self):
        return self.args['repo']

    def get_packages(self, downloader, list_or_file_path, property_validate=True):
        return self._adapter.get_packages(self, self._parser, downloader, list_or_file_path, property_validate)

    def get_usedby(self, downloader, list_or_file_path, property_validate=True):
        return self._adapter.get_usedby(self, self._parser, downloader, list_or_file_path, property_validate)

    def __getattr__(self, item):
        return self.args.get(item, None)

    def __getitem__(self, item):
        return self.args.get(item, None)

    def generate_full_urls_from_package_path_in_repo(self, package_path_in_repo):
        res = []

        for repo in self.repos:
            res.append("{}/{}/{}".format(self.args.server, repo, package_path_in_repo))

        return res

    def get_paths(self, packages_to_find):
        paths = []

        for package, repo, path_pattern in itertools.product(packages_to_find, self.repos, self.path_patterns):
            params = {
                'server': self.args['server'],
                'repo': repo,
                'package': package['package'],
                'version': package['version']
            }
            paths.append(Dict({'path': path_pattern.format(**params), 'params': params}))

        return paths

    def get_auth_params(self):
        art_auth_etc = Dict()
        art_auth_etc.auth = HTTPBasicAuth(*tuple(self.args['auth']))

        if 'verify' in self.args:
            art_auth_etc.verify = self.args.verify.lower in ['true', 'yes', '1']
        else:
            art_auth_etc.verify = False
        return art_auth_etc



class DepsTxtLockListFormatter:
    def __init__(self):
        pass

    @staticmethod
    def write_packages_to_lock_file(deps_lock_file_path, packages):

        with open(deps_lock_file_path, 'w+') as f:
            for p in packages.values():
                print(p.art_package.art_path.path_in_repo, file=f)

    @staticmethod
    def read_packages_from_lock_file(deps_lock_file_path):

        with open(deps_lock_file_path) as f:
            return f.read().splitlines()



import re

from packaging.version import Version

CROSSPM_VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\-(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""

class PackageVersion(Version):

    _regex = re.compile(r"^\s*" + CROSSPM_VERSION_PATTERN + r"\s*$", re.VERBOSE | re.IGNORECASE)

    def __init__(self, version):
        super(PackageVersion, self).__init__(version)
        self._version_original_str = version

    # cant use base Version.__str__, because 'local' placeholder normalized, and all '[-_]' are changed with '.'
    def __str__(self):
        return self._version_original_str



from tabulate import tabulate

class DepsTxtLockTableFormatter:
    def __init__(self):
        pass

    @staticmethod
    def write(lockfile_full_path, packages):

        packages_for_table = [[p.art_package.name, p.art_package.version] for p in packages.values()]
        table = tabulate(packages_for_table, headers=[], tablefmt='plain')

        with open(lockfile_full_path, 'w+') as f:
            f.write(table)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from setuptools import setup
from crosspm import config

build_number = os.getenv("TRAVIS_BUILD_NUMBER", "")
branch = os.getenv("TRAVIS_BRANCH", "")
travis = any((build_number, branch,))
version = config.__version__.split(".")
develop_status = "4 - Beta"
url = "https://devopshq.github.io/crosspm2/"

if travis:
    version = version[0:3]
    if branch == "master":
        develop_status = "5 - Production/Stable"
        version.append(build_number)
    else:
        version.append("{}{}".format("dev" if branch == "develop" else branch, build_number))
else:
    if len(version) < 4:
        version.append("dev0")

version = ".".join(version)
if travis:
    with open("crosspm/config.py", "w", encoding="utf-8") as f:
        f.write("__version__ = '{}'".format(version))

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="crosspm2",
    version=version,
    description="Cross Package Manager 2",
    license="MIT",
    author="Timut Gilmullin",
    author_email="tim55667757@gmail.com",
    url=url,
    long_description=long_description,
    download_url="https://github.com/devopshq/crosspm2.git",
    entry_points={"console_scripts": ["crosspm2=crosspm.__main__:main"]},
    classifiers=[
        "Development Status :: {}".format(develop_status),
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    keywords=[
        "development",
        "dependency",
        "requirements",
        "manager",
        "versioning",
        "packet",
    ],
    packages=[
        "crosspm",
        "crosspm.helpers",
        "crosspm.adapters",
        "crosspm.template",
        "crosspm.contracts",
        "crosspm.deps_txt_parsers",
        "crosspm.output_formatters",
        "crosspm.package_parsers"
    ],
    setup_requires=[
        "wheel==0.34.2",
        "pypandoc==1.5",
    ],
    tests_require=[
        "flake8==3.7.9",
        "pytest<=4.6.9; python_version < '3.5'",
        "pytest>=5.2; python_version >= '3.5'",
        "pytest-flask<1.0.0; python_version < '3.5'",
        "pytest-flask>=1.0.0; python_version >= '3.5'",
        "PyYAML<5.2; python_version < '3.5'",
        "PyYAML>=5.2; python_version >= '3.5'",
    ],
    install_requires=[
        "requests<2.22; python_version < '3.5'",
        "requests>=2.22; python_version >= '3.5'",
        "urllib3==1.24.3",
        "docopt==0.6.2",
        "PyYAML==5.1.2; python_version < '3.5'",
        "PyYAML>=5.2; python_version >= '3.5'",
        "dohq-artifactory==0.7.377",
        "dohq-artifactory==0.4.112; python_version < '3.5'",
        "dohq-artifactory>=0.7.377; python_version >= '3.5'",
        "Jinja2<2.11; python_version < '3.5'",
        "Jinja2>=2.11; python_version >= '3.5'",
        "patool==1.12",  # need for pyunpack
        "pyunpack==0.1.2",
        "addict==2.2.1",  # need for crosspm2 contract scheme
        "packaging>=20.4",  # need for crosspm2 contract scheme
        "parse>=1.15.0",  # need for crosspm2 contract scheme
        "tabulate>=0.8.7",  # need for crosspm2 contract scheme
        "ordered-set>=4.0.2",

    ],
    package_data={
        "": [
            "./template/*.j2",
            "*.cmake",
            "../LICENSE",
            "../README.*",
        ],
    },
)



from parse import compile

from crosspm.helpers.exceptions import CrosspmException, CROSSPM_ERRORCODE_VERSION_PATTERN_NOT_MATCH

DEBIAN_PACKAGENAME_PATTERN = compile('{}_{}_{}.deb')

# https://www.debian.org/doc/manuals/debian-faq/pkg-basics.en.html
# 7.3. Why are Debian package file names so long?
# The Debian binary package file names conform to the following convention:
# <foo>_<VersionNumber>-<DebianRevisionNumber>_<DebianArchitecture>.deb


class DebianPackageNameParser:

    def __init__(self, package, version, revision, arch):
        self.package = package
        self.version = version
        self.revision = revision
        self.arch = arch

    @classmethod
    def parse_from_package_name(cls, package_name):

        try:
            package, fullversion, arch = DEBIAN_PACKAGENAME_PATTERN.parse(package_name)
            version, sep, revision = fullversion.partition('-')
            return DebianPackageNameParser(package, version, revision, arch)
        except TypeError:
            raise CrosspmException(CROSSPM_ERRORCODE_VERSION_PATTERN_NOT_MATCH, f"package name <{package_name}> mismatch debian name convension pattern <foo>_<Version>_<DebianArchitecture>.deb")


    def __str__(self):
        return self.fullname

    def __repr__(self):
        return str(self)

    @property
    def fullname(self):
        return "{}_{}_{}.deb".format(self.package, self.fullversion, self.arch)

    @property
    def fullversion(self):

        if self.revision:
            return f"{self.version}-{self.revision}"
        return f"{self.version}"



# -*- coding: utf-8 -*-
import http.client
import json
import logging
from collections import OrderedDict

import requests
from addict import Dict
from artifactory import ArtifactoryPath

import crosspm.contracts.package
from crosspm.adapters.artifactoryaql import ArtifactoryAql
from crosspm.contracts.bundle import Bundle
from crosspm.helpers.exceptions import *  # noqa
from crosspm.helpers.package import Package

PACKAGE_PROPERTY_CONTRACT_PREFFIX = 'c.'
CHUNK_SIZE = 1024

setup = {
    "name": [
        "jfrog-artifactory-aql2",
    ],
}

session = requests.Session()


class ArtifactoryAql2(ArtifactoryAql):
    def get_packages(self, source, parser, downloader, list_or_file_path, property_validate=True):
        """

        :param source:
        :param parser:
        :param downloader:
        :param list_or_file_path:
        :param property_validate: for `root` packages we need check property, bad if we find packages from `lock` file,
        we can skip validate part
        :return:
        """
        # logging.basicConfig(level=logging.DEBUG)
        # http.client.HTTPConnection.debuglevel = 1
        # requests_log = logging.getLogger("requests.packages.urllib3")
        # requests_log.setLevel(logging.DEBUG)
        # requests_log.propagate = True

        _art_auth_etc = source.get_auth_params()

        _pkg_name_column = self._config.name_column
        _secret_variables = self._config.secret_variables
        _packages_found = OrderedDict()
        _packed_exist = False
        _packed_cache_params = None
        self._log.info('parser: {}'.format(parser._name))

        repo_returned_packages_all = []

        for _path in source.get_paths(list_or_file_path['raw']):

            last_error = ''

            _tmp_params = Dict(_path.params)
            self._log.info('repo: {}'.format(_tmp_params.repo))

            _path_fixed, _path_pattern, _file_name_pattern = parser.split_fixed_pattern_with_file_name(_path.path)
            _package_versions_with_contracts = self.find_package_versions(_art_auth_etc, _file_name_pattern,
                                                                          _path_pattern, _tmp_params,
                                                                          last_error,
                                                                          parser, _tmp_params)

            _package = None

            if _package_versions_with_contracts:
                repo_returned_packages_all += _package_versions_with_contracts

        package_names = [x[self._config.name_column] for x in list_or_file_path['raw']]

        bundle = Bundle(package_names, repo_returned_packages_all,
                        downloader._config.trigger_package)

        bundle_packages = bundle.calculate().values()
        for p in bundle_packages:
            _packages_found[p.name] = Package(p.name, p.art_path, p, p.paths_params, downloader, self, parser,
                                              p.params, p.params_raw, p.pkg_stat())

        for p in package_names:
            if p not in _packages_found.keys():
                _packages_found[p.name] = None

        return _packages_found

    def parse_contracts_from_items_find_results(self, properties):
        contracts = Dict()

        for p in properties:
            if p.key.startswith(PACKAGE_PROPERTY_CONTRACT_PREFFIX):
                contracts[p.key] = p.value[0]

        return contracts

    def find_package_versions(self, _art_auth_etc, _file_name_pattern,
                              _path_pattern, _tmp_params, last_error, parser, paths_params):
        try:
            package_versions_with_contracts = []
            _artifactory_server = _tmp_params.server
            _search_repo = _tmp_params.repo

            # Get AQL path pattern, with fixed part path, without artifactory url and repository name
            _aql_path_pattern = ""
            if _path_pattern:
                _aql_path_pattern = _path_pattern

            _aql_query_url = '{}/api/search/aql'.format(_artifactory_server)
            _aql_query_dict = {
                "repo": {
                    "$eq": _search_repo,
                },
                "path": {
                    "$match": _aql_path_pattern,
                },
                "name": {
                    "$match": _file_name_pattern,
                },
            }
            # Remove path if is empty string
            if not _aql_path_pattern:
                _aql_query_dict.pop('path')
            query = 'items.find({query_dict}).include("*", "property")'.format(
                query_dict=json.dumps(_aql_query_dict))
            session.auth = _art_auth_etc['auth']
            r = session.post(_aql_query_url, data=query, verify=_art_auth_etc['verify'])
            r.raise_for_status()

            _found_paths = Dict(r.json())
            for _found in _found_paths.results:
                _repo_path = "{artifactory}/{repo}/{path}/{file_name}".format(
                    artifactory=_artifactory_server,
                    repo=_found.repo,
                    path=_found.path,
                    file_name=_found.name)
                _repo_path = ArtifactoryPath(_repo_path, **_art_auth_etc)

                _mark = 'found'
                _matched, _params, _params_raw = parser.validate_path(str(_repo_path), _tmp_params)
                self._log.info('_matched:{}, _params:{}, _params_raw:{}, _repo_path:{}'.format(_matched, _params, _params_raw, str(_repo_path)))

                if _matched:
                    contracts = self.parse_contracts_from_items_find_results(_found.properties)

                    package_with_contracts = crosspm.contracts.package.ArtifactoryPackage(_params.package,
                                                                                          _params_raw.version,
                                                                                          contracts, _repo_path,
                                                                                          _params, _params_raw,
                                                                                          paths_params,
                                                                                          _found
                                                                                          )

                    package_versions_with_contracts.append(package_with_contracts)

                    _mark = 'valid'

                self._log.info('  {}: {}'.format(_mark, str(_repo_path)))
        except RuntimeError as e:
            try:
                err = json.loads(e.args[0])
            except Exception:
                err = {}
            if isinstance(err, dict):
                # Check errors
                # :e.args[0]: {
                #                 "errors" : [ {
                #                     "status" : 404,
                #                     "message" : "Not Found"
                #                  } ]
                #             }
                for error in err.get('errors', []):
                    err_status = error.get('status', -1)
                    err_msg = error.get('message', '')
                    if err_status == 401:
                        msg = 'Authentication error[{}]{}'.format(err_status,
                                                                  (': {}'.format(
                                                                      err_msg)) if err_msg else '')
                    elif err_status == 404:
                        msg = last_error
                    else:
                        msg = 'Error[{}]{}'.format(err_status,
                                                   (': {}'.format(err_msg)) if err_msg else '')
                    if last_error != msg:
                        self._log.error(msg)
                    last_error = msg
        return package_versions_with_contracts

    # def find_package_in_repo_by_fullname(self, source, package_fullname):
    #
    #     return

    def download_package(self, source, package_fullname, package_local_fullpath):
        packages = self.find_package_versions(self.get_auth(source), package_fullname, '', {}, last_error, self.parser,
                                              {})

        with packages[0].art_path.open() as fd, open(package_local_fullpath, "wb") as out:
            out.write(fd.read())

